#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdint.h>

/*#define DEBUG 1*/

#pragma pack(push, 1)
typedef struct {
  uint32_t magic;
  uint8_t unknown[28];
} mov_head;

typedef struct {
  uint8_t  cID;
  uint32_t cSize;
} mov_part;

typedef struct {
  uint8_t  IDLength;
  uint8_t  ColorMapType;
  uint8_t  ImageType;
  uint16_t CMapStart;
  uint16_t CMapLength;
  uint8_t  CMapDepth;
  uint16_t XOffset;
  uint16_t YOffset;
  uint16_t Width;
  uint16_t Height;
  uint8_t  PixelDepth;
  uint8_t  ImageDescriptor;
} tga_head;

typedef struct {
  uint32_t dwRIFFSign;
  uint32_t dwRIFFSize;
  uint32_t dwWAVESign;
  uint32_t dwfmt_Sign;
  uint32_t dwfmt_Size;
  /* PCMWAVEFORMAT */
  /* WAVEFORMAT */
  uint16_t wFormatTag;
  uint16_t nChannels;
  uint32_t nSamplesPerSec;
  uint32_t nAvgBytesPerSec;
  uint16_t nBlockAlign;
  /* /WAVEFORMAT */
  uint16_t wBitsPerSample;
  /* /PCMWAVEFORMAT */
  uint32_t dwdataSign;
  uint32_t dwdataSize;
} wav_head;
#pragma pack(pop)

#define BUF_SIZE (320 * 200)
static uint8_t buf[BUF_SIZE];
static uint8_t img[BUF_SIZE];
static uint8_t pal[256][3];
static uint32_t f_num;

void FillWaveHeader(wav_head *hdr, uint32_t len, uint32_t sps, uint16_t chn, uint16_t bps) {
  if (hdr) {
    memset(hdr, 0, sizeof(hdr[0]));
    hdr->dwRIFFSign = 0x46464952;
    hdr->dwRIFFSize = sizeof(hdr[0]) - 8 + len;
    hdr->dwWAVESign = 0x45564157;
    hdr->dwfmt_Sign = 0x20746D66;
    hdr->dwfmt_Size = 16;
    hdr->wFormatTag = 1;
    hdr->nChannels  = chn;
    hdr->nSamplesPerSec = sps;
    hdr->wBitsPerSample = bps;
    hdr->nBlockAlign = hdr->nChannels * (hdr->wBitsPerSample / 8);
    hdr->nAvgBytesPerSec = hdr->nBlockAlign * hdr->nSamplesPerSec;
    hdr->dwdataSign = 0x61746164;
    hdr->dwdataSize = len;
  }
}

void readdata(FILE *fl, void *data, uint32_t size) {
  if (data && size) {
    memset(data, 0, size);
    if (fl) {
      fread(data, size, 1, fl);
    }
  }
}

void dumpthis(void) {
uint32_t x, y;
tga_head th;
#ifndef DEBUG
char s[32];
FILE *f;
#endif
  memset(&th, 0, sizeof(th));
  th.Width = 320;
  th.Height = 200;
  th.ColorMapType = 1;
  th.ImageType = 1;
  th.CMapLength = 256;
  th.CMapDepth = 24;
  th.PixelDepth = 8;
  th.ImageDescriptor = 0x20;

  /* rotate image 90* */
  for (y = 0; y < 200; y++) {
    for (x = 0; x < 320; x++) {
      img[(y * 320) + x] = buf[(x * 200) + y];
    }
  }

#ifndef DEBUG
  sprintf(s, "%08u.tga", f_num);
  f = fopen(s, "wb");
  if (f) {
    fwrite(&th, sizeof(th), 1, f);
    fwrite(pal, sizeof(pal), 1, f);
    fwrite(img, sizeof(img), 1, f);
    fclose(f);
  }
#endif

  f_num++;
}

int main(int argc, char *argv[]) {
uint8_t a, b, *p;
uint32_t flsize;
mov_head mh;
mov_part mc;
wav_head wh;
uint16_t i;
FILE *fl, *fa;
  printf(
    "Exhumed / PowerSlave video file BOOK.MOV decoder v1.1\n"
    "(c) -=CHE@TER=- 2010,2012,2024\n"
    "http://ctpax-cheater.losthost.org/htmldocs/ps_vs_ex.htm\n\n"
  );
  fl = fopen("BOOK.MOV", "rb");
  if (!fl) {
    printf("Error: can not open file \"BOOK.MOV\" for read.\n\n");
    return(1);
  }
  readdata(fl, &mh, sizeof(mh));
  /* "LMF " */
  if (mh.magic != 0x20464D4C) {
    fclose(fl);
    printf("Error: unknown or invalid format for \"BOOK.MOV\" file.\n\n");
    return(2);
  }
  fseek(fl, 0, SEEK_END);
  flsize = ftell(fl);
  fseek(fl, sizeof(mh), SEEK_SET);
  /* init */
  fa = fopen("BOOK.WAV", "wb");
  if (!fa) {
    fclose(fl);
    printf("Error: can not create output audio file \"BOOK.WAV\" for write.\n\n");
    return(3);
  }
  memset(&wh, 0, sizeof(wh));
  fwrite(&wh, sizeof(wh), 1, fa);
  f_num = 0;
  /* decode */
  while (ftell(fl) < flsize) {
    readdata(fl, &mc, sizeof(mc));
    switch (mc.cID - 1) {
      case 0: /* PALETTE */
        readdata(fl, pal, sizeof(pal));
        mc.cSize -= sizeof(pal);
        readdata(fl, &b, 1);
        mc.cSize--; /* here mc.cSize must be zero */
        memset(buf, b, BUF_SIZE); /* fill buffer with initial color */
        for (i = 0; i < 256; i++) {
          /* VGA [0..63] -> RGB [0..255] */
          for (a = 0; a < 3; a++) {
            pal[i][a] = ((pal[i][a] << 2) | (pal[i][a] >> 4));
          }
          /* RGB -> BGR */
          pal[i][0] ^= pal[i][2];
          pal[i][2] ^= pal[i][0];
          pal[i][0] ^= pal[i][2];
        }
        break;
      case 1: /* AUDIO */
        p = (uint8_t *) malloc(mc.cSize);
        if (p) {
          readdata(fl, p, mc.cSize);
          fwrite(p, 1, mc.cSize, fa);
          free(p);
        }
        break;
      case 2: /* VIDEO */
        readdata(fl, &i, 2);
        mc.cSize -= 2;
        p = buf; /* image start offs */
        p += (200 * i);
        while (mc.cSize > 0) {
          readdata(fl, &a, 1);
          readdata(fl, &b, 1);
          mc.cSize -= 2;
          p += a;
          if (b) {
            readdata(fl, p, b);
            mc.cSize -= b;
            p += b;
          }
        }
        dumpthis();
        break;
      case 3: /* SYNC */
        fseek(fl, mc.cSize, SEEK_CUR);
        dumpthis();
        break;
      default: /* UNKNOWN */
        fseek(fl, mc.cSize, SEEK_CUR);
        break;
    }
  }
  /* or 22222 ? or maybe this stored in .MOV header? */
  FillWaveHeader(&wh, ftell(fa) - sizeof(wh), 22050, 1, 8);
  /* fix wave header */
  fseek(fa, 0, SEEK_SET);
  fwrite(&wh, sizeof(wh), 1, fa);
  fclose(fa);
  fclose(fl);
  return(0);
}
